"""
Scraipe is a web scraping and analysis framework.
Check out https://github.com/snpm/scraipe for more information.
"""
from scraipe.workflow import Workflow
from scraipe.classes import IScraper, IAnalyzer, ScrapeResult, AnalysisResult